"""
    MrTopo Network - data structure repr networks
"""

from mininet.topolib import TreeTopo, TorusTopo
from mininet.topo import LinearTopo, MinimalTopo
from mrtopo.structures.networks.ring.ring import RingTopo
from mrtopo.structures import Topology
import copy


class Network:
    def __init__(self):
        self.raw = None  # dict lines
        self.topology = None  # used by mutator to generate networks following given topology's rules
        self.switches = []
        self.links = []
        self.hosts = []
        self.network = {}
        self.host_network = {}

    def init_network(self):
        for name in self.switches:
            self.network[name] = []

    def init_host_network(self):
        for name in self.hosts:
            self.host_network[name] = []

    def link_exists(self, s_1, s_2):
        return [s_1, s_2] in self.links

    def switch_exists(self, s):
        return s in self.switches

    def host_exists(self, h):
        return h in self.hosts

    def add_switch(self, s):
        if not self.switch_exists(s):
            host = s + "_host"
            self.switches.append(s)
            self.hosts.append(host)
            self.host_network[host] = [s]
            self.network[s] = [host]

    def remove_switch(self, s):
        if self.switch_exists(s):
            snet = None
            if s.__contains__("host"):
                snet = self.host_network
                self.hosts.remove(s)
            else:
                snet = self.network
                self.switches.remove(s)
            # TODO REMOVE FROM LINKS
            for key in list(snet.keys()):
                if snet[key].__contains__(s):
                    snet[key].remove(s)

    def add_link(self, s1, s2):
        if self.switch_exists(s1) and self.switch_exists(s2):
            if not self.link_exists(s1, s2):
                snet = None
                if s1.__contains__("host") or s2.__contains__("host"):
                    snet = self.host_network
                else:
                    snet = self.network

                self.add(s1, s2)
                self.add(s2, s1)
                self.links.append([s1, s2])
                self.links.append([s2, s1])

    def remove_link(self, s1: str, s2: str):
        if self.link_exists(s1, s2):
            snet = None
            if s1.__contains__("host") or s2.__contains__("host"):
                snet = self.host_network
            else:
                snet = self.network

            self.remove(s1, s2)
            self.remove(s2, s1)
            self.link_remove(s1, s2)

    def deep_copy(self):
        deep_copy = Network()
        deep_copy.raw = copy.deepcopy(self.raw)
        deep_copy.topology = copy.deepcopy(self.topology)
        deep_copy.switches = copy.deepcopy(self.switches)
        deep_copy.links = copy.deepcopy(self.links)
        deep_copy.hosts = copy.deepcopy(self.hosts)
        deep_copy.network = copy.deepcopy(self.network)
        deep_copy.host_network = copy.deepcopy(self.host_network)
        return deep_copy

    def remove(self, _from, switch):
        if _from.__contains__("host"):
            self.host_network[_from].remove(switch)
        else:
            self.network[_from].remove(switch)

    def link_remove(self, s1, s2):
        if [s1, s2] in self.links:
            self.links.remove([s1, s2])
        if [s2, s1] in self.links:
            self.links.remove([s2, s1])

    def add(self, _from, switch):
        if _from.__contains__("host"):
            self.host_network[_from].append(switch)
        else:
            self.network[_from].append(switch)

    def get_switches(self):
        return self.switches

    def get_hosts(self):
        return self.hosts

    def get_links(self):
        return self.links

    def __repr__(self):
        return "<Network Topology( " + str(self.topology.__class__) + " )>"

    def __eq__(self, other):
        return self.topology.__class__ == other.topology.__class__

    @staticmethod
    def build_links(links, n):

        network = n.network
        host_network = n.host_network
        for link in links:
            line = link["line_str"]
            start = line.find("(")

            var_count = 0
            _link = ["", ""]
            for i in range(start + 1, len(line) - 1):
                if var_count == 2:
                    break  # TODO add options later
                elif line[i] == ",":
                    var_count += 1
                else:
                    _link[var_count] += line[i]

            for i in range(len(_link)):
                _link[i] = _link[i].strip()

            for l in _link:
                _snet = None
                if l.__contains__("host"):
                    _snet = host_network
                else:
                    _snet = network

                for _l in _link:
                    if _l != l:
                        _snet[l].append(_l)

            n.links.append(_link)  # TODO SUPPORT LINKS FOR ALL

    @staticmethod
    def p_build(lines: dict):
        n = Network()
        n.raw = lines

        keys = list(lines.keys())

        n.switches = get_var_names(lines["switch_lines"])

        n.hosts = get_var_names(lines["host_lines"])

        n.init_network()
        n.init_host_network()

        Network.build_links(lines["link_lines"], n)

        return n

    @staticmethod
    def c_build(topology: Topology):
        n = Network()
        # TODO MODULARIZE
        if topology == Topology.DEFAULT:
            pass  # TODO
        elif topology == Topology.MINIMAL:
            n.topology = MinimalTopo()
        elif topology == Topology.LINEAR:
            n.topology = LinearTopo()
        elif topology == Topology.RING:
            n.topology = RingTopo()
        elif topology == Topology.TREE:
            n.topology = TreeTopo()
        elif topology == Topology.TORUS:
            n.topology = TorusTopo()
        return n


def get_var_names(coll):
    names = []
    for item in coll:
        name = ""
        for c in item['line_str']:
            if c == '=':
                break
            else:
                name += str(c)
        names.append(name.strip())
    return names



import sys, getopt
from mrtopo.logger import log
from mrtopo.structures import config, Topology
from mrtopo.mutator import mutate
from mrtopo.translator import c_read, p_read, m_write, desc_write
from mrtopo.interpreter import interpret
from mrtopo.filetype import FileType
import docopt

# from . import __version__ as version

__version__ = "0.0.1"


def options(argv: list):
    try:
        opts, args = getopt.getopt(argv, "hc:p", ["configfile="])
    except getopt.GetoptError:
        print('Usage: mrtopo -c <configfile>')
        sys.exit(2)
    for opt, arg in opts:
        if opt == '-h':
            print('usage: mrtopo:  -c <configfile>\n\t\t-p <pyfile>')
            sys.exit()
        elif opt in ("-p", "--pyfile"):
            return (args[0], FileType.PYTHON)
        elif opt in ("-c", "--configfile"):
            return (args[0], FileType.CONFIG)


def main(argv):

    log('MrTopo.v.' + __version__ + '>')

    _file, file_type = options(argv)

    if file_type == FileType.CONFIG:
        configuration = config.Config()  # initialize default config

        # translate
        if _file:
            configuration = c_read(_file)

        # interpret
        interpret(configuration)

    elif file_type == FileType.PYTHON:

        pyfile = None

        # translate
        if _file:
            pyfile = p_read(_file)

        network = None
        if pyfile:
            # interpret
            network = interpret(pyfile)

        mutant_networks = mutate(network)

        for mutant in mutant_networks:
            m_write(mutant, _file)

        desc_write(mutant_networks)


if __name__ == "__main__":
    main(sys.argv[1:])



"""
    MrTopo Translator - translates user input to interpretable data
"""

from mrtopo.logger import log
from mrtopo.error import ErrorIdentities, handle_error
from mrtopo.translator.mrtopoio import find_file, read_json, open_read
from mrtopo.mutator.operators import Operations
from mrtopo.structures import config, mutantnetwork, network
from shutil import copyfile
from datetime import date
import os


# Read a config file
def c_read(file: str):
    log('Translator - Reading: ' + file)
    _j = find_file(file)
    _d = None

    if len(_j) == 1:
        _d = read_json(_j[0])
    else:
        pass
        # TODO ERROR
    if _d:
        configuration = config.Config.build(_d)
    else:
        configuration = config.Config()

    return configuration  # TODO Validation


# Read a py file
def p_read(file: str):  # TODO ADD TESTS
    log('Translator - Reading: ' + file)
    _p = find_file(file)
    _d = {
        "switch_lines": [],
        "link_lines": [],
        "host_lines": []
    }
    o_rf = open_read(file)
    line_no = 0
    for l in o_rf:
        if l.__contains__("self.addSwitch"):
            _d["switch_lines"].append(to_line_obj(l, line_no))
        if l.__contains__("self.addHost"):
            _d["host_lines"].append(to_line_obj(l, line_no))
        if l.__contains__("self.addLink"):
            _d["link_lines"].append(to_line_obj(l, line_no))
        line_no += 1

    o_rf.close()
    return _d


def to_line_obj(line, number):
    return {"line_str": line.strip(), "line_no": number}


def make_folder():
    _dir = os.path.dirname(os.path.realpath(__file__))
    dir_name = "MrTopoGenerated"
    path = os.path.join(_dir, dir_name)
    if not os.path.isdir("MrTopoGenerated"):
        os.mkdir("MrTopoGenerated")
    return "MrTopoGenerated"


def m_write(mutant: mutantnetwork.MutantNetwork, original_file: str):
    filename = "mrtopo_gen_topo_" + str(mutant._id) + ".py"

    path = make_folder() + "/" + filename

    copyfile(original_file, path)

    operation = mutant.get_operation()
    modified = mutant.get_modified_item()

    read_file = open(path, "r")
    line_list = read_file.readlines()

    read_file.close()

    n_switches = len(mutant.network.get_switches())
    n_hosts = len(mutant.network.get_hosts())

    if operation == Operations.ADDSWITCH:
        new_switch = modified[0]
        link_switch = modified[1]
        switch_string = add_switch_string(new_switch, n_switches + 1)
        host_string = add_host_string(new_switch + "_host", n_hosts + 1)
        host_link_string = add_host_link_string(new_switch + "_host", new_switch)
        switch_link_string = add_link_string(new_switch, link_switch)

        for i in range(len(line_list)):
            if line_list[i].__contains__("self.addSwitch"):
                whitespace_len = len(line_list[i]) - len(line_list[i].strip()) - 1
                line_list.insert(i, ' ' * whitespace_len + switch_string)
                line_list.insert(i + 1, ' ' * whitespace_len + host_string)
                break

        for i in range(len(line_list)):
            if line_list[i].__contains__("self.addLink"):
                whitespace_len = len(line_list[i]) - len(line_list[i].strip()) - 1
                line_list.insert(i, ' ' * whitespace_len + host_link_string)
                line_list.insert(i + 1, ' ' * whitespace_len + switch_link_string)  # add host and new switch link
                break

    if operation == Operations.ADDLINK:
        link_string = add_link_string(modified[0], modified[1])
        for i in range(len(line_list)):
            if line_list[i].__contains__("self.addLink"):
                whitespace_len = len(line_list[i]) - len(line_list[i].strip()) - 1
                line_list.insert(i, ' ' * whitespace_len + link_string)
                break

    if operation == Operations.REMOVELINK:
        for i in range(len(line_list)):
            if line_list[i].__contains__(modified[0]) and line_list[i].__contains__(modified[1]) \
                    and line_list[i].__contains__("self.addLink"):
                line_list[i] = "# " + line_list[i]
                break

    if operation == Operations.REMOVESWITCH:
        for i in range(len(line_list)):
            if line_list[i].__contains__(modified):
                line_list[i] = "# " + line_list[i]

    line_list.insert(0, "# MRTOPO MODIFICATION DESCRIPTION:\n# " + mutant.get_description() + "\n")

    write_file = open(path, "w")

    for line in line_list:
        write_file.write(line)

    write_file.close()


def desc_write(mutants: mutantnetwork.MutantNetwork, outfile="desc.txt"):  # create description doc for all cases

    path = make_folder() + "/" + outfile

    out_lines = []
    today = date.today().strftime("%d/%m/%Y")
    out_lines.append("MrTopo network mutation description file.\n")
    out_lines.append("Generated on {}\n{}\n".format(today, '-' * 10))

    for m in mutants:
        out_lines.append("Case {}: {}\n".format(m.get_id(), m.get_operation()))
        out_lines.append("{}\n".format(m.get_description()))
        out_lines.append("Modified item(s): {}\n".format(m.get_modified_item()))
        out_lines.append("{}\n".format('-' * 10))

    write_file = open(path, "w")

    for line in out_lines:
        write_file.write(line)

    write_file.close()


def add_switch_string(name, number):
    return name + " = " + "self.addSwitch(" + "\'s" + str(number) + "\' ) # MRTOPO GENERATED LINE\n"


def add_host_string(name, number):
    return name + " = " + "self.addHost(" + "\'h" + str(number) + "\' ) # MRTOPO GENERATED LINE\n"


def add_link_string(s1, s2):
    return "self.addLink({}, {}, bw=10, delay='0.345064487693ms') # MRTOPO GENERATED LINE\n".format(s1,s2)
    # TODO randomize opts


def add_host_link_string(s1, s2):
    return "self.addLink({}, {}) # MRTOPO GENERATED LINE\n".format(s1, s2)  # TODO randomize opts



import setuptools
import re

with open("README.md", "r") as fh:
    long_description = fh.read()

version = re.search(
    '^__version__\s*=\s*"(.*)"',
    open('mrtopo/__main__.py').read(),
    re.M
    ).group(1)

setuptools.setup(
    name='mrtopo',
    version=version,
    packages=setuptools.find_packages(),
    url='https://github.com/FaizChishtie/mrtopo',
    license='MIT',
    author='faizchishtie',
    author_email='faizchishtie@gmail.com',
    description='Mutate Mininet topology files with MrTopo',
    python_requires='>=3.0',
    # entry_points={'console_scripts': ['mrtopo = mrtopo.__main__:main']},
    long_description=long_description,
    install_requires=[
            'mininet',
            'docopt'
    ]
)



"""
    MrTopo Topology - enum type for topologies
"""

from enum import Enum


# Enum for topologies
class Topology(Enum):
    DEFAULT = 0
    MINIMAL = 1
    LINEAR = 2
    TREE = 3
    RING = 4
    TORUS = 5



"""
    MrTopo Config - data structure for config files
"""
from mrtopo.structures.topology import Topology
from enum import Enum
from mrtopo.logger import log
import json


# Following https://wiki.onosproject.org/display/ONOS/Test+Plans
class TestPlan(Enum):
    DEFAULT = 0
    # CORE
    CHO = 1
    HA = 2
    FUNC = 3
    SPCF = 4
    PLAT = 5
    SAMP = 6
    USECASE = 7

    def __str__(self):
        return self.name


def find_enum(name, _enum):
    for item in _enum:
        if name == item.name:
            return item
    return _enum.DEFAULT  # default must exist for enums passed in TODO


# Configuration for system
class Config:
    def __init__(self):
        self.plan = TestPlan.DEFAULT
        self.topologies = [TestPlan.DEFAULT]

    def __eq__(self, other):
        return other.plan == self.plan and other.topologies == self.topologies

    def __repr__(self):
        top = ""
        for t in self.topologies:
            top += str(t) + " "
        return "<Config Plan( " + str(self.plan) + " ) Topologies( " + top + ")>"

    # build config structure from json
    @staticmethod
    def build(_dict: dict):
        c = Config()

        if "plan" in _dict:
            c.plan = find_enum(_dict["plan"], TestPlan)

        if "topologies" in _dict:
            c.topologies.clear()
            topologies = _dict["topologies"]
            if type(topologies) is list:
                for t in topologies:
                    c.topologies.append(find_enum(t, Topology))
            elif type(topologies) is str:
                c.topologies.append(find_enum(topologies, Topology))
        return c
